import socket
import json
import threading
import time
import signal
import queue
import logging
import sys
import DAO


class TCPClient(threading.Thread):
    """
    对于每一个机台的Tcp客户端，用于接收数据的线程
    :param ip:机台ip地址
    :param port:机台端口
    :param name:机台名称
    """
    def __init__(self, ip, port,name):
        threading.Thread.__init__(self)
        self.ip = ip
        self.port = port
        self.stopflag = False
        self.isconnected = False
        self.equipname=name
        self.logger = logging.getLogger(self.equipname+"-thread")

    def run(self):
        while not self.stopflag: # 循环接收数据
            if not self.isconnected: # 先判断连接状态
                self.connect()
                if not self.isconnected:
                    continue
            try:
                data = b''
                while True: # 拼接数据，如果收到数据0.5秒内没有再收到数据，则认为接收结束
                    try:
                        buff = self.sock.recv(1024)
                        if len(buff) == 0: # 某些情况下，收到数据为空也是断开
                            raise Exception("连接已断开")
                        data = data + buff
                    except socket.timeout:
                        if data:
                            break
                data = data.decode('utf-8',errors='ignore')
                if not q.full():
                    q.put(data) # 解码后放入队列
            except Exception as e:
                self.logger.error(f"连接中断:{e}")
                self.isconnected = False
                self.sock.close()
        self.logger.debug("停止循环接收数据")
    
    def connect(self):
        """
        连接tcp服务器
        """
        while not self.isconnected and not self.stopflag:
            try:
                self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.sock.settimeout(3) # 3秒内没连接成功视为连接失败
                self.sock.connect((self.ip, self.port))
                self.sock.settimeout(0.5)
                self.isconnected = True
                self.logger.info("连接成功")
            except Exception as e:
                self.logger.info(f"重连失败{e}")
                self.sock.close()
                time.sleep(2)
    
    def stop(self):
        """
        停止线程
        """
        self.stopflag = True
        self.sock.close()


class DataHandler(threading.Thread):
    """
    解析并存储数据的线程
    """
    def __init__(self):
        threading.Thread.__init__(self)
        self.stopflag = False
        self.logger = logging.getLogger("datahandler")

    def run(self):
        while not self.stopflag:
            json_data = q.get() # 等待队列数据
            if json_data is None: # 如果队列中元素是None，代表停止处理，程序退出
                q.task_done()
                break
            TotalOutPutList = []
            NGDetailsList = []
            NGHourDetailsList = []
            try:
                data = json.loads(json_data)
                # 解析总产量json
                for TotalOutPut in data["TotalOutPut"]:# 遍历JSON对象
                    twoCameraFlag = False
                    for item in TotalOutPutList:# 查找是否有已存在的数据，并将两个相机的机台汇总到一起
                        if item["StatisticsDate"] == TotalOutPut["StatisticsDate"] and item["ShiftNum"] == TotalOutPut["ShiftNum"] and item["ProductNo"] == TotalOutPut["ProductNo"] and item["ParentEquipNo"] == TotalOutPut["ParentEquipNo"]:
                            item["AllNum"] = item["AllNum"] + TotalOutPut["AllNum"]
                            item["NGNum"] = item["NGNum"] + TotalOutPut["NGNum"]
                            item["OKNum"] = item["OKNum"] + TotalOutPut["OKNum"]
                            twoCameraFlag = True
                    if twoCameraFlag == False:# 如果没有该机台数据，则新增
                        TotalOutPutList.append({
                            "AllNum":TotalOutPut["AllNum"],
                            "NGNum":TotalOutPut["NGNum"],
                            "OKNum":TotalOutPut["OKNum"],
                            "ParentEquipNo":TotalOutPut["ParentEquipNo"],
                            "ProductName":TotalOutPut["ProductName"],
                            "ProductNo":TotalOutPut["ProductNo"],
                            "ShiftNum":TotalOutPut["ShiftNum"],
                            "StatisticsDate":TotalOutPut["StatisticsDate"]
                        })

                # 解析NG明细json
                for NGDetails in data["NGDetails"]:# 遍历JSON对象
                    twoCameraFlag = False
                    for item in NGDetailsList:# 查找是否有已存在的数据，并将两个相机的机台汇总到一起
                        if item["StatisticsDate"] == NGDetails["F_StatisticsDate"] and item["ShiftNum"] == NGDetails["ShiftNum"] and item["ProductNo"] == NGDetails["ProductNo"] and item["ParentEquipNo"] == NGDetails["ParentEquipNo"] and item["NGClassNo"] == NGDetails["NGClassNo"]:
                            item["Num"] = item["Num"] + NGDetails["Num"]
                            twoCameraFlag = True
                    if twoCameraFlag == False:# 如果没有该机台数据，则新增
                        NGDetailsList.append({
                            "EquipNo":NGDetails["EquipNo"],
                            "Num":NGDetails["Num"],
                            "NGClassNo":NGDetails["NGClassNo"],
                            "NGClassName":NGDetails["NGClassName"],
                            "ParentEquipNo":NGDetails["ParentEquipNo"],
                            "ProductName":NGDetails["ProductName"],
                            "ProductNo":NGDetails["ProductNo"],
                            "ShiftNum":NGDetails["ShiftNum"],
                            "StatisticsDate":NGDetails["F_StatisticsDate"]
                        })

                # 解析ng小时分布
                for NGHourDetails in data["NGHourDetails"]:# 遍历JSON对象
                    twoCameraFlag = False
                    for item in NGHourDetailsList:# 查找是否有已存在的数据，并将两个相机的机台汇总到一起
                        if item["StatisticsDate"] == NGHourDetails["F_StatisticsDate"] and item["ShiftNum"] == NGHourDetails["ShiftNum"] and item["ProductNo"] == NGHourDetails["ProductNo"] and item["ParentEquipNo"] == NGHourDetails["ParentEquipNo"] and item["NGClassNo"] == NGHourDetails["NGClassNo"] and item["CreateDate"] == NGHourDetails["CreateDate"]:
                            item["Num"] = item["Num"] + NGHourDetails["Num"]
                            twoCameraFlag = True
                    if twoCameraFlag == False:# 如果没有该机台数据，则新增
                        NGHourDetailsList.append({
                            "EquipNo":NGHourDetails["EquipNo"],
                            "Num":NGHourDetails["Num"],
                            "CreateDate":NGHourDetails["CreateDate"],
                            "NGClassNo":NGHourDetails["NGClassNo"],
                            "NGClassName":NGHourDetails["NGClassName"],
                            "ParentEquipNo":NGHourDetails["ParentEquipNo"],
                            "ProductName":NGHourDetails["ProductName"],
                            "ProductNo":NGHourDetails["ProductNo"],
                            "ShiftNum":NGHourDetails["ShiftNum"],
                            "StatisticsDate":NGHourDetails["F_StatisticsDate"]
                        })
            except: # 如果解析过程的任何失败都认为json解析失败，并丢弃当前数据
                self.logger.error(f"json解析失败:{json_data}")
                q.task_done()
                continue
            self.logger.debug("解析TotalOutPutList:")
            self.logger.debug(TotalOutPutList)
            self.logger.debug("解析NGDetailsList:")
            self.logger.debug(NGDetailsList)
            self.logger.debug("解析NGHourDetailsList:")
            self.logger.debug(NGHourDetailsList)
            # 如果解析成功，依次存入数据库
            db.saveTotalOutPut(TotalOutPutList)
            db.saveNGDetails(NGDetailsList)
            db.saveNGHourDetails(NGHourDetailsList)
            # 标记队列中该数据已处理
            q.task_done()

    def stop(self):
        q.put(None)
        self.stopflag = True


def quit_handler(signum, frame): # 退出程序
    global quitflag
    rootlogger.debug(f"退出中")
    for TcpClient in TcpClientList:
        rootlogger.debug(f"退出{TcpClient.equipname}")
        TcpClient.stop()
        TcpClient.join()
    rootlogger.debug(f"退出datahandler")
    while not q.empty(): # 清空队列
        try:
            item = q.get_nowait()
            q.task_done()
        except queue.Empty:
            break
    datahandler.stop()
    datahandler.join()
    rootlogger.debug(f"关闭数据库连接")
    db.close()
    quitflag = True
    print("正常退出")

# 配置logging
stream_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
stream_handler.setFormatter(formatter)

rootlogger = logging.getLogger()
rootlogger.addHandler(stream_handler)


# 初始化
TcpClientList = []
try:
    conf = json.load(open("conf.json", "r",encoding="utf-8"))

    logLevel = conf['logLevel'] # 配置文件三个级别
    if logLevel == "debug":
        rootlogger.setLevel(logging.DEBUG)
    elif logLevel == "info":
        rootlogger.setLevel(logging.INFO)
    elif logLevel == "warning":
        rootlogger.setLevel(logging.WARNING)
    else:
        rootlogger.setLevel(logging.ERROR)
    dbHost = conf['dataBase']['host']
    dbPort = conf['dataBase']['port']
    dbUserName = conf['dataBase']['username']
    dbPassword = conf['dataBase']['password']
    dbDatabase = conf['dataBase']['Database']
    machineList = conf['machineList']
    for machine in machineList:
        TcpClientList.append(TCPClient(machine["host"], machine["port"],machine["name"]))

except Exception as e:
    rootlogger.fatal(f"初始化失败{e}")
    sys.exit(1)

q = queue.Queue(maxsize=1000) # 待解析数据队列
db = DAO.MySQLDB(dbHost,dbPort,dbUserName,dbPassword,dbDatabase)# 初始化数据库

quitflag = False
signal.signal(signal.SIGINT,quit_handler) # 注册中断信号处理函数

for TcpClient in TcpClientList:# 启动全部tcp客户端
    TcpClient.start()

datahandler = DataHandler()
datahandler.start()

while not quitflag:
    time.sleep(1)

